// AddPrinter.cpp:  Console utility to install printers


#define UNICODE			1
#define _UNICODE		1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <winspool.h>

#include "tcpxcv.h"


#define VER_MAJOR		1
#define VER_MINOR		3
#define VER_BUILD		0
#define VER_PATCH		1



const wchar_t AddPrinterSyntaxText[] = L"Install a new local printer.\n\
\n\
ADDPRINTER printername drivername [/A /C /D /I /L /O /P /Q /R /T]\n\
\n\
\t/A:address    host address\n\
\t/C:string     printer comment\n\
\t/D            make the new printer the default\n\
\t/I:filename   INF filename\n\
\t/L:string     printer location\n\
\t/O            only install if it doesn't already exist\n\
\t/P:port       port name\n\
\t/Q:queue      LPR queue name\n\
\t/R:filename   read printer settings from a file\n\
\t/T            send a test page\n\
\n\
Specify either /A or /P but not both (/A automatically creates an\n\
IP port.)  The filename may contain <Win> or <x64> tags to select\n\
the appropriate driver for the target system.  Quote any parameter\n\
containing spaces or other special characters.\n";



const wchar_t DefaultQueueName[]	= L"raw";


const wchar_t AddPrinterRegKey[]	= L"SOFTWARE\\University of New Mexico\\College of Education\\AddPrinter";
const wchar_t DisabledRegValue[]	= L"Disable";



#define SHORT_BUF_LEN				1024
#define MAX_ARG_LEN					8192



#ifdef _WIN64

#define PLATFORM L"x64"

#else	//	_WIN64

#define PLATFORM L"x86"

#endif	//	_WIN64



int DisplayErrorHeader()
{
	fwprintf( stderr, L"ADDPRINTER:  " );

	return 0;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( !FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) )
		wsprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



long int GetInt( LPTSTR inString )
{
	if ( !inString )
		return -1;

	long int rv = 0;
	wchar_t *sep1 = NULL;

	if ( !iswdigit( inString[0] ) )
		return -1;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) )
		rv = wcstol( inString + 2, &sep1, 16 );
	else
		rv = wcstol( inString, &sep1, 10 );

	if ( errno )
		return -1;
	else if ( *sep1 != '\0' )
		return -1;

	return rv;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



int ConsoleShow( LPTSTR outString )
{
	DWORD ConsoleMode	= 0;


	if ( !GetConsoleMode( GetStdHandle( STD_OUTPUT_HANDLE ), &ConsoleMode ) )
		return 0;

	wprintf( L"\r                                                                      \r" );

	if ( outString )
		wprintf( L"%s", outString );

	return 0;
}



int IsWin64()
{

#ifdef _WIN64

	//  If app is compiled for 64-bit Windows, then duh....
	return 1;

#else	//	_WIN64

	//  App compiled for Win32; need to check whether it's actually running on Win64.

	wchar_t dirbuf[MAX_ARG_LEN]		= L"";

	typedef UINT ( __stdcall *GET64DIRPROC )( LPTSTR, UINT );
	HMODULE hKernel32		= GetModuleHandle( L"kernel32.dll" );
	GET64DIRPROC Get64Dir	= NULL;

	if ( !hKernel32 )
		return 0;

	Get64Dir = (GET64DIRPROC) GetProcAddress( hKernel32, "GetSystemWow64DirectoryW" );

	if ( !Get64Dir )
		return 0;

	if ( (Get64Dir) ( dirbuf, MAX_ARG_LEN ) == 0 )
		return 0;

	if ( !iswalpha( dirbuf[0] ) )
		return 0;
	
	if ( dirbuf[1] != ':' )
		return 0;

	if ( dirbuf[2] != '\\' )
		return 0;
	else
		return 1;

#endif	//	_WIN64
}



int GetWindowsVersion( LPTSTR WindowsVer )
{
	if ( !WindowsVer )
		return -1;


	OSVERSIONINFOEXW VerInfo;
	typedef NTSTATUS ( WINAPI _RtlGetVersionProc ) ( OSVERSIONINFOEXW *VerInfo );
	HMODULE NtDllHandle						= GetModuleHandle( L"NtDll.dll" );
	_RtlGetVersionProc *RtlGetVersionProc	= NULL;


	memset( &VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo.dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	if ( NtDllHandle )
		RtlGetVersionProc = (_RtlGetVersionProc*) GetProcAddress( NtDllHandle, "RtlGetVersion" );

	if ( RtlGetVersionProc )
		RtlGetVersionProc( &VerInfo );

	if ( VerInfo.dwMajorVersion == 0 )
		GetVersionExW( (OSVERSIONINFOW*) &VerInfo );


	if ( ( VerInfo.dwMajorVersion == 5 ) && ( VerInfo.dwMinorVersion == 0 ) )
		wcscpy_s( WindowsVer, SHORT_BUF_LEN, L"Win2K" );
	else if ( ( VerInfo.dwMajorVersion == 5 ) && ( VerInfo.dwMinorVersion >= 1 ) && ( VerInfo.dwMinorVersion <= 2 ) )
		wcscpy_s( WindowsVer, SHORT_BUF_LEN, L"XP" );
	else if ( ( VerInfo.dwMajorVersion == 6 ) && ( VerInfo.dwMinorVersion == 0 ) )
		wcscpy_s( WindowsVer, SHORT_BUF_LEN, L"Vista" );
	else if ( ( VerInfo.dwMajorVersion == 6 ) && ( VerInfo.dwMinorVersion == 1 ) )
		wcscpy_s( WindowsVer, SHORT_BUF_LEN, L"Win7" );
	else if ( ( VerInfo.dwMajorVersion == 6 ) && ( VerInfo.dwMinorVersion >= 2 ) )
		wcscpy_s( WindowsVer, SHORT_BUF_LEN, L"Win8" );
	else if ( VerInfo.dwMajorVersion == 10 )
		wcscpy_s( WindowsVer, SHORT_BUF_LEN, L"Win10" );
	else
		wsprintf( WindowsVer, L"Win%d.%d", VerInfo.dwMajorVersion, VerInfo.dwMinorVersion );


	/*
	if ( ( VerInfo.dwMajorVersion > 5 ) || ( ( VerInfo.dwMajorVersion == 5 ) && ( VerInfo.dwMinorVersion >= 1 ) ) )
		if ( IsWin64() )
			wcscat_s( WindowsVer, SHORT_BUF_LEN, L"-x64" );
	*/

	//	wprintf( L"   * GetWindowsVersion() returning '%s'.\n", WindowsVer );
	return 0;
}



int ParseFilenameTag( LPTSTR inString, int *index )
{
	if ( !inString )
		return 0;


	wchar_t shortbuf[SHORT_BUF_LEN]		= L"";
	int i		= 0;
	wchar_t ch	= '\0';


	do {
		ch = inString[i];
		shortbuf[i++] = towupper( ch );
	} while ( ( ch != '>' ) && ( ch != '\0' ) && ( i < (SHORT_BUF_LEN - 1) ) );

	if ( ch != '>' )
		return 0;

	shortbuf[i - 1] = '\0';

	if ( !_wcsicmp( shortbuf, L"64" ) || !_wcsicmp( shortbuf, L"x64" ) ) {
		*index += i;
		return 1;
	}
	else if ( !_wcsicmp( shortbuf, L"win" ) || !_wcsicmp( shortbuf, L"ver" ) ) {
		*index += i;
		return 2;
	}

	return 0;
}



int InsertOSVersion( LPTSTR Filename )
{
	if ( !Filename )
		return -1;

	wchar_t WindowsVer[SHORT_BUF_LEN] = L"";
	GetWindowsVersion( WindowsVer );

	BOOL IsX64 = IsWin64();

	wchar_t Temp[MAX_ARG_LEN] = L"";
	int i = 0;
	size_t j = 0;
	wchar_t ch = '\0';

	do {
		ch = Filename[i++];

		if ( ch == '<' ) {

			int tag = ParseFilenameTag( Filename + i, &i );

			switch ( tag ) {

				case 1:			// <x64>
					if ( IsX64 ) {
						wcscat_s( Temp, MAX_ARG_LEN, L"-x64" );
						j = wcslen( Temp );
					}

					break;

				case 2:			// <win>
					wcscat_s( Temp, MAX_ARG_LEN, WindowsVer );
					j = wcslen( Temp );
					break;

				default:
					Temp[j] = ch;
					if ( j < ( MAX_ARG_LEN - 1 ) )
						j++;

					Temp[j] = '\0';

			}
		}
		else {
			Temp[j] = ch;

			if ( j < ( MAX_ARG_LEN - 1 ) )
				j++;

			Temp[j] = '\0';
		}


	}
	while ( ch );

	wcscpy_s(Filename, MAX_ARG_LEN, Temp );
	return 0;
}



int ReplaceChar( LPTSTR inString, wchar_t fromChar, wchar_t toChar )
{
	if ( !inString )
		return -1;

	int i = 0;
	wchar_t ch = '\0';

	do {
		if ( ( ch = inString[i] ) == fromChar )
			ch = toChar;

		inString[i++] = ch;
	
	} while ( ch != '\0' );

	return 0;
}



BOOL LooksLikeAnIP( LPTSTR inString )
{
	if ( !inString )
		return FALSE;

	int octet = 0, i = 0;
	int val = 0;

	do {
		if ( !iswdigit( inString[i] ) )
			return FALSE;
		else
			val = inString[i++] - '0';

		if ( iswdigit( inString[i] ) )
			val = ( val * 10 ) + ( inString[i++] - '0' );

		if ( iswdigit( inString[i] ) )
			val = ( val * 10 ) + ( inString[i++] - '0' );

		if ( inString[i++] != ( ( octet == 3 ) ? '\0' : '.' ) )
			return FALSE;

		if ( val > 255 )
			return FALSE;

		octet++;

	} while ( octet < 4 );

	return TRUE;
}



int GeneratePortName( LPTSTR HostAddress, LPTSTR PortName )
{
	if ( !HostAddress || !PortName )
		return -1;

	wchar_t shortbuf[SHORT_BUF_LEN] = L"IP__";

	if ( !LooksLikeAnIP( HostAddress ) ) {

		size_t i = 0, j = wcslen( shortbuf );
		wchar_t ch = '\0';

		do {
			ch = HostAddress[i++];

			if ( j < ( SHORT_BUF_LEN - 1 ) ) {
				shortbuf[j++] = towupper( ch );
				shortbuf[j] = '\0';
			}

		} while ( ch != '\0' );

		if ( j < ( SHORT_BUF_LEN - 1 ) ) {
			shortbuf[j - 1] = '$';
			shortbuf[j] = '\0';
		}

		wcscpy_s( PortName, SHORT_BUF_LEN, shortbuf );
		return 0;
	}
	

	int val = 0, hi = 0, lo = 0;
	size_t i = 0, j = wcslen( shortbuf );
	wchar_t ch = '\0', hdig = '\0', ldig = '\0' ;

	do {
		ch = HostAddress[i++];

		if ( iswdigit( ch ) )
			val = ( 10 * val ) + ( ch - '0' );
		else {
			hi = val / 16;
			lo = val & 15;

			if ( hi < 10 )
				hdig = '0' + hi;
			else
				hdig = 'A' + hi - 10;

			if ( lo < 10 )
				ldig = '0' + lo;
			else
				ldig = 'A' + lo - 10;

			shortbuf[j++] = hdig;
			shortbuf[j++] = ldig;

			val = 0;
		}

	} while ( ch != '\0' );

	shortbuf[j++] = '$';
	shortbuf[j] = '\0';

	wcscpy_s( PortName, SHORT_BUF_LEN, shortbuf );
	return 0;
}



BOOL PortNameInList( LPTSTR Name, PORT_INFO_1 *PortNames, unsigned int NumPorts )
{
	if ( !Name || !PortNames || !NumPorts )
		return FALSE;


	for ( unsigned int i = 0; i < NumPorts; i++ )
		if ( !_wcsicmp( Name, PortNames[i].pName ) )
			return TRUE;

	return FALSE;
}



int CreateIPPort( LPTSTR PortName, LPTSTR HostName, LPTSTR IPAddress, LPTSTR QueueName, int Protocol, int PortNumber )
{
	//		Check inputs :

	if ( !PortName )
		return -1;

	if ( wcslen( PortName ) >= MAX_PORTNAME_LEN )
		return -1;

	if ( !HostName && !IPAddress )
		return -1;



	wchar_t temp[SHORT_BUF_LEN]		= L"";

	PORT_INFO_1 *PortNames;
	DWORD BytesNeeded			= 0;
	DWORD NumPorts				= 0;

	size_t len					= 0;
	int err						= 0;
	int j						= 0;
	wchar_t *sep1				= NULL;

	PORT_DATA_1 PortData;
	HANDLE hXcv					= INVALID_HANDLE_VALUE;
	PRINTER_DEFAULTS Defaults	= { NULL, NULL, SERVER_ACCESS_ADMINISTER };
	DWORD XcvStatus				= 0;


	//		Get list of existing ports :

	EnumPorts( NULL, 1, NULL, 0, &BytesNeeded, &NumPorts );

	if ( BytesNeeded ) {
		PortNames = (PORT_INFO_1*) malloc( BytesNeeded );

		if ( !PortNames ) {
			err = ERROR_NOT_ENOUGH_MEMORY;
			goto ErrorExit;
		}

		if ( !EnumPorts( NULL, 1, (BYTE*) PortNames, BytesNeeded, &BytesNeeded, &NumPorts ) )
			goto ErrorExit;
	}


	//		Generate default port name, if none was supplied :

	if ( HostName )
		if ( HostName[0] )
			GeneratePortName( HostName, temp );

	if ( temp[0] == '\0' )
		if ( IPAddress )
			if ( IPAddress[0] )
				GeneratePortName( IPAddress, temp );

	if ( PortName[0] == '\0' )
		wcscpy_s( PortName, SHORT_BUF_LEN, temp );

	sep1 = wcsrchr( temp, '$' );
	if ( sep1 )
		*sep1 = '\0';


	//		If port name matches an existing port, generate a unique name for the new port :

	while ( PortNameInList( PortName, PortNames, NumPorts ) )
		wsprintf( PortName, L"%s_%u$", temp, ++j );


	//		Check host and IP address name lengths :

	if ( HostName ) {
		len += wcslen( HostName );

		if ( wcslen( HostName ) >= MAX_NETWORKNAME_LEN )
			goto ErrorExit;
	}

	if ( IPAddress ) {
		len += wcslen( IPAddress );

		if ( wcslen( IPAddress ) >= MAX_IPADDR_STR_LEN )
			goto ErrorExit;
	}

	if ( !len )
		goto ErrorExit;


	//		Sanity check protocol and TCP/IP port :

	if ( !Protocol )
		Protocol = PROTOCOL_RAWTCP_TYPE;

	if ( ( Protocol != PROTOCOL_RAWTCP_TYPE ) && ( Protocol != PROTOCOL_LPR_TYPE ) )
		goto ErrorExit;

	if ( !PortNumber ) {
		if ( Protocol == PROTOCOL_RAWTCP_TYPE )
			PortNumber = 9100;
		else
			PortNumber = 515;
	}


	//		Set up PORT_INFO_1 structure for creation of a new IP printing port :

	memset( &PortData, 0, sizeof( PORT_DATA_1 ) );

	wcscpy_s( PortData.sztPortName, MAX_PORTNAME_LEN, PortName );
	PortData.dwVersion		= 1;
	PortData.dwProtocol		= Protocol;
	PortData.cbSize			= sizeof( PortData );
	PortData.dwPortNumber	= PortNumber;
	PortData.dwSNMPDevIndex	= 1;
	wcscpy_s( PortData.sztSNMPCommunity, MAX_SNMP_COMMUNITY_STR_LEN, L"public" );

	if ( !HostName )
		wcscpy_s( PortData.sztHostAddress, MAX_NETWORKNAME_LEN, IPAddress );
	else if ( HostName[0] == '\0' )
		wcscpy_s( PortData.sztHostAddress, MAX_NETWORKNAME_LEN, IPAddress );
	else
		wcscpy_s( PortData.sztHostAddress, MAX_NETWORKNAME_LEN, HostName );

	if ( IPAddress )
		wcscpy_s( PortData.sztIPAddress, MAX_IPADDR_STR_LEN, IPAddress );

	if ( Protocol == PROTOCOL_LPR_TYPE ) {
		if ( !QueueName )
			wsprintf( PortData.sztQueue, DefaultQueueName );
		else if ( QueueName[0] == '\0' )
			wsprintf( PortData.sztQueue, DefaultQueueName );
		else
			wcscpy_s( PortData.sztQueue, MAX_QUEUENAME_LEN, QueueName );
	}


	//		Create a new IP port:

	if ( OpenPrinter( L",XcvMonitor Standard TCP/IP Port", &hXcv, &Defaults ) ) {

		if ( !XcvData( hXcv, L"AddPort", (PBYTE) &PortData, sizeof( PortData ), NULL, 0, &BytesNeeded, &XcvStatus ) ) {
			err = GetLastError();
			goto ErrorExit;
		}

		ClosePrinter( hXcv );
	}
	else {
		err = GetLastError();
		goto ErrorExit;
	}


	//		Success!

	free( PortNames );
	return 0;


ErrorExit:

	//		Failure!

	if ( !err )
		err = -1;

	if ( hXcv && ( hXcv != INVALID_HANDLE_VALUE ) )
		ClosePrinter( hXcv );

	if ( PortNames)
		free( PortNames );
	
	return err;
}



BOOL PrinterExists( LPTSTR PrinterName )
{
	HANDLE PrinterHandle = NULL;
	DWORD rv = 0;

	if ( OpenPrinter( PrinterName, &PrinterHandle, NULL ) ) {
		ClosePrinter( PrinterHandle );
		return TRUE;
	}

	rv = GetLastError();

	if ( rv == ERROR_INVALID_PRINTER_NAME ) 
		return FALSE;
	
	return TRUE;
}



int IsNetPrinterName( LPTSTR inString )
{
	if ( !inString )
		return 0;

	if ( inString[0] != '\\' )
		return 0;

	if ( inString[1] != '\\' )
		return 0;

	if ( ( inString[2] == '\\' ) || ( inString[2] == '\0' ) )
		return 0;

	int i = 2;
	while ( ( inString[i] != '\\' ) && ( inString[i] != '\0' ) )
		i++;

	if ( inString[i] != '\\' )
		return 0;

	if ( ( inString[i + 1] == '\\' ) || ( inString[i] == '\0' ) )
		return 0;

	return 1;
}



int TweakPortName( LPTSTR PortName )
{
	if ( !PortName )
		return -1;
	else if ( PortName[0] == '\0' )
		return -1;

	BOOL knownport = FALSE;

	knownport |= !_wcsnicmp( PortName, L"LPT", 3 );
	knownport |= !_wcsnicmp( PortName, L"COM", 3 );

	if ( !knownport )
		return 0;

	size_t i = 3;
	if ( !iswdigit( PortName[i] ) )
		return 0;

	while ( iswdigit( PortName[i] ) )
		i++;

	if ( ( PortName[i] == '\0' ) || ( ( PortName[i] == ':' ) && ( PortName[i + 1] == '\0' ) ) )
		_wcsupr_s( PortName, MAX_ARG_LEN );

	if ( PortName[i] )
		return 0;

	if ( i < ( SHORT_BUF_LEN - 1 ) ) {
		PortName[i++] = ':';
		PortName[i] = '\0';
	}

	return 0;
}



BOOL CheckDisableRegistryKeys()
{
	HKEY KeyHandle		= NULL;
	BOOL disabled		= FALSE;
	DWORD disabledval	= 0;
	DWORD buflen		= 4;


	if ( RegOpenKeyEx( HKEY_CURRENT_USER, AddPrinterRegKey, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {

		buflen = 4;
		disabledval = 0;

		if ( RegQueryValueEx( KeyHandle, DisabledRegValue, NULL, NULL, (LPBYTE) &disabledval, &buflen ) == ERROR_SUCCESS )
			disabled = ( disabledval != 0 );

		RegCloseKey( KeyHandle );
	}

	if ( disabled ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"Disabled by user registry key.\n" );
		return TRUE;
	}

	if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, AddPrinterRegKey, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {

		buflen = 4;
		disabledval = 0;

		if ( RegQueryValueEx( KeyHandle, DisabledRegValue, NULL, NULL, (LPBYTE) &disabledval, &buflen ) == ERROR_SUCCESS )
			disabled = ( disabledval != 0 );

		RegCloseKey( KeyHandle );
	}

	if ( disabled ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"Disabled by system registry key.\n" );
		return TRUE;
	}

	return FALSE;
}



BOOL IsAFile( LPTSTR Filename )
{
	if ( !Filename )
		return FALSE;
	else if ( Filename[0] == '\0' )
		return FALSE;

	DWORD Attribs = GetFileAttributes( Filename );

	if ( Attribs == INVALID_FILE_ATTRIBUTES )
		return FALSE;

	if ( Attribs & ( FILE_ATTRIBUTE_DEVICE | FILE_ATTRIBUTE_DIRECTORY ) )
		return FALSE;

	return TRUE;
}



int SetPrinterPrefs( LPTSTR PrinterName, int Color, int Duplex, int Orientation, int PaperSize, BOOL Complain )
{
	if ( !PrinterName )
		return -1;
	else if ( PrinterName[0] == '\0' )
		return -1;

	
	HANDLE PrinterHandle				= NULL;
	DEVMODE *Defaults					= NULL;
	int ChangeLevel						= 9;
	PRINTER_INFO_9 NewDefaults;
	DWORD BytesNeeded					= 0;
	wchar_t ErrorText[SHORT_BUF_LEN]	= L"";

	int rv								= 0;


	if ( OpenPrinter( PrinterName, &PrinterHandle, NULL ) ) {

		BytesNeeded = DocumentProperties( NULL, PrinterHandle, PrinterName, NULL, NULL, 0 );

		if ( BytesNeeded ) {
			Defaults = (DEVMODE*) malloc( BytesNeeded );

			if ( Defaults ) {

				if ( DocumentProperties( NULL, PrinterHandle, PrinterName, Defaults, NULL, DM_OUT_BUFFER ) == IDOK ) {

					Defaults->dmFields = 0;

					if ( Color )
						Defaults->dmColor = Color, Defaults->dmFields |= DM_COLOR;

					if ( Duplex )
						Defaults->dmDuplex = Duplex, Defaults->dmFields |= DM_DUPLEX;

					if ( Orientation )
						Defaults->dmOrientation = Orientation, Defaults->dmFields |= DM_ORIENTATION;

					if ( PaperSize )
						Defaults->dmPaperSize = PaperSize, Defaults->dmFields |= DM_PAPERSIZE;

					NewDefaults.pDevMode = Defaults;

					if ( SetPrinter( PrinterHandle, ChangeLevel, (BYTE*) &NewDefaults, 0 ) )
						rv = 0;
					else {
						//	SetPrinter() failed:

						rv = 5;

						if ( Complain ) {
							GetErrorText( GetLastError(), ErrorText );
							DisplayErrorHeader();
							fwprintf( stderr, L"Failed to set defaults for \"%s\"\n", PrinterName );
							fwprintf( stderr, L"   %s\n", ErrorText );
						}
					}
				}
				else {
					//	DocumentProperties() failed:

					rv = 4;

					if ( Complain ) {
						DisplayErrorHeader();
						fwprintf( stderr, L"Error getting DEVMODE structure\n" );
					}
				}

				free( Defaults );
			}
			else {
				//	malloc() failed:

				rv = 3;

				if ( Complain ) {
					DisplayErrorHeader();
					fwprintf( stderr, L"Can't allocate DEVMODE structure\n" );
				}
			}
		}
		else {
			//	DocumentProperties() returned 0:

			rv = 2;

			if ( Complain ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Can't get DEVMODE structure size\n" );
			}
		}

		ClosePrinter( PrinterHandle );
	}
	else {
		//	error opening printer:

		rv = 1;

		if ( Complain ) {
			GetErrorText( GetLastError(), ErrorText );
			DisplayErrorHeader();
			fwprintf( stderr, L"Error opening printer \"%s\":\n", PrinterName );
			fwprintf( stderr, L"   %s\n", ErrorText );
		}
	}


	return rv;
}



int wmain( int argc, wchar_t *argv[] )
{
	
	wchar_t PrinterName[SHORT_BUF_LEN]		= L"";
	wchar_t DriverName[SHORT_BUF_LEN]		= L"";
	wchar_t PortName[SHORT_BUF_LEN]			= L"";
	wchar_t HostAddress[SHORT_BUF_LEN]		= L"";
	wchar_t LPRQueueName[SHORT_BUF_LEN]		= L"";
	wchar_t InfFilename[MAX_ARG_LEN]		= L"";
	wchar_t PtrComment[SHORT_BUF_LEN]		= L"";
	wchar_t PtrLocation[SHORT_BUF_LEN]		= L"";
	wchar_t DataFilename[MAX_ARG_LEN]		= L"";
	wchar_t ErrorMessage[SHORT_BUF_LEN]		= L"";

	WORD TCPPort = 9100;

	BOOL TestPage				= FALSE;
	BOOL MakeDefault			= FALSE;
	BOOL AdditionalSettings		= FALSE;
	BOOL NoOverwrite			= FALSE;
	BOOL BeVerbose				= FALSE;
	BOOL DoNothing				= FALSE;
	BOOL TCPPortSpecified		= FALSE;
	BOOL HostNameIsDottedQuad	= FALSE;
	BOOL LPRPrinterPort			= FALSE;

	int PrefColor				= 0;
	int PrefDuplex				= 0;
	int PrefOrient				= 0;
	int PrefPaper				= 0;
	BOOL ChangePrefs			= FALSE;

	int rv						= 0;


	int i = 1;

	while ( i < argc ) {

		if ( wcslen( argv[i] ) >= MAX_ARG_LEN ) {
			DisplayErrorHeader();
			fwprintf( stderr, L"Argument #%d is way too long!\n", i );
			return 1;
		}


		if ( !_wcsicmp( argv[i], L"/?" ) || !_wcsicmp( argv[i], L"/??" ) ) {
			_putws( AddPrinterSyntaxText );
			wprintf( L"\t\t\t\tAddPrinter.exe  v%d.%d.%d (%s)\n", VER_MAJOR, VER_MINOR, VER_BUILD, PLATFORM );
			return 0;
		}
		else if ( !_wcsicmp( argv[i], L"/T" ) )
			TestPage = TRUE;
		else if ( !_wcsicmp( argv[i], L"/D" ) )
			MakeDefault = TRUE;
		else if ( !_wcsicmp( argv[i], L"/O" ) )
			NoOverwrite = TRUE;
		else if ( !_wcsicmp( argv[i], L"/$" ) )
			BeVerbose = TRUE;
		else if ( !_wcsicmp( argv[i], L"/$$" ) )
			BeVerbose = DoNothing = TRUE;
		else if ( !_wcsnicmp( argv[i], L"/A", 2 ) ) {
			//  Get host address:

			int j = 2;

			while ( iswalpha( argv[i][j] ) )
				j++;

			if ( IsColonOrEquals( argv[i][j] ) )
				j++;
			
			if ( argv[i][j] == '\0' ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Syntax error: \"%s\"\n", argv[i] );
				return 1;
			}

			wcscpy_s( HostAddress, SHORT_BUF_LEN, argv[i] + j );
		}
		else if ( !_wcsnicmp( argv[i], L"/P", 2 ) ) {
			//  Get port name:

			int j = 2;

			while ( iswalpha( argv[i][j] ) )
				j++;

			if ( IsColonOrEquals( argv[i][j] ) )
				j++;
			
			if ( argv[i][j] == '\0' ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Syntax error: \"%s\"\n", argv[i] );
				return 1;
			}

			wcscpy_s( PortName, SHORT_BUF_LEN, argv[i] + j );
		}
		else if ( !_wcsnicmp( argv[i], L"/Q", 2 ) ) {
			//  Get LPR queue name:

			int j = 2;

			while ( iswalpha( argv[i][j] ) )
				j++;

			if ( IsColonOrEquals( argv[i][j] ) )
				j++;;

			if ( argv[i][j] == '\0' )
				wcscpy_s( LPRQueueName, SHORT_BUF_LEN, DefaultQueueName );
			else
				wcscpy_s( LPRQueueName, SHORT_BUF_LEN, argv[i] + j );
		}
		else if ( !_wcsnicmp( argv[i], L"/I", 2 ) ) {
			//  Get INF filename:

			int j = 2;

			while ( iswalpha( argv[i][j] ) )
				j++;

			if ( IsColonOrEquals( argv[i][j] ) )
				j++;
			
			if ( argv[i][j] == '\0' ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Syntax error: \"%s\"\n", argv[i] );
				return 1;
			}

			wcscpy_s( InfFilename, MAX_ARG_LEN, argv[i] + j );
		}
		else if ( !_wcsnicmp( argv[i], L"/C", 2 ) ) {
			//  Get printer comment:

			int j = 2;

			while ( iswalpha( argv[i][j] ) )
				j++;

			if ( IsColonOrEquals( argv[i][j] ) )
				j++;
			
			if ( argv[i][j] == '\0' ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Syntax error: \"%s\"\n", argv[i] );
				return 1;
			}

			wcscpy_s( PtrComment, SHORT_BUF_LEN, argv[i] + j );
			AdditionalSettings = TRUE;
		}
		else if ( !_wcsnicmp( argv[i], L"/L", 2 ) ) {
			//  Get printer location:

			int j = 2;

			while ( iswalpha( argv[i][j] ) )
				j++;

			if ( IsColonOrEquals( argv[i][j] ) )
				j++;
			
			if ( argv[i][j] == '\0' ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Syntax error: \"%s\"\n", argv[i] );
				return 1;
			}

			wcscpy_s( PtrLocation, SHORT_BUF_LEN, argv[i] + j );
			AdditionalSettings = TRUE;
		}
		else if ( !_wcsnicmp( argv[i], L"/R", 2 ) ) {
			//  Get printer data filename:

			int j = 2;

			while ( iswalpha( argv[i][j] ) )
				j++;

			if ( IsColonOrEquals( argv[i][j] ) )
				j++;

			if ( argv[i][j] == '\0' ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Missing filename: \"%s\"\n", argv[i] );
				return 1;
			}

			wcscpy_s( DataFilename, MAX_ARG_LEN, argv[i] + j );
		}
		else if ( !_wcsnicmp( argv[i], L"/XC", 3 ) ) {
			//	Change default color setting:

			int j = 3;

			if ( IsColonOrEquals( argv[i][j] ) )
				j++;

			PrefColor = GetInt( argv[i] + j );

			if ( PrefColor < 0 ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Bad value: \"%s\"\n", argv[i] + j );
				return 1;
			}
			else if ( ( PrefColor < DMCOLOR_MONOCHROME ) || ( PrefColor > DMCOLOR_COLOR ) ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Default color must be %u to %u: %u\n", DMCOLOR_MONOCHROME, DMCOLOR_COLOR, PrefColor );
				return 1;
			}

			ChangePrefs = TRUE;
		}
		else if ( !_wcsnicmp( argv[i], L"/XD", 3 ) ) {
			//	Change default duplex setting:

			int j = 3;

			if ( IsColonOrEquals( argv[i][j] ) )
				j++;

			PrefDuplex = GetInt( argv[i] + j );

			if ( PrefDuplex < 0 ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Bad value: \"%s\"\n", argv[i] + j );
				return 1;
			}
			else if ( ( PrefDuplex < DMDUP_SIMPLEX ) || ( PrefDuplex > DMDUP_HORIZONTAL ) ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Default duplex must be %u to %u: %u\n", DMDUP_SIMPLEX, DMDUP_HORIZONTAL, PrefDuplex );
				return 1;
			}

			ChangePrefs = TRUE;
		}
		else if ( !_wcsnicmp( argv[i], L"/XO", 3 ) ) {
			//	Change default orientation setting:

			int j = 3;

			if ( IsColonOrEquals( argv[i][j] ) )
				j++;

			PrefOrient = GetInt( argv[i] + j );

			if ( PrefOrient < 0 ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Bad value: \"%s\"\n", argv[i] + j );
				return 1;
			}
			else if ( ( PrefOrient < DMORIENT_PORTRAIT ) || ( PrefOrient > DMORIENT_LANDSCAPE ) ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Default orientation must be %u to %u: %u\n", DMORIENT_PORTRAIT, DMORIENT_LANDSCAPE, PrefOrient );
				return 1;
			}

			ChangePrefs = TRUE;
		}
		else if ( !_wcsnicmp( argv[i], L"/XP", 3 ) ) {
			//	Change default paper size setting:

			int j = 3;

			if ( IsColonOrEquals( argv[i][j] ) )
				j++;

			PrefPaper = GetInt( argv[i] + j );

			if ( PrefPaper < 0 ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Bad value: \"%s\"\n", argv[i] + j );
				return 1;
			}
			else if ( ( PrefPaper < DMPAPER_LETTER ) || ( PrefPaper > DMPAPER_LAST ) ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Default paper size must be %u to %u: %u\n", DMPAPER_LETTER, DMPAPER_LAST, PrefPaper );
				return 1;
			}

			ChangePrefs = TRUE;
		}
		else if ( ( argv[i][0] != '/' ) && ( PrinterName[0] == '\0' ) )
			wcscpy_s( PrinterName, SHORT_BUF_LEN, argv[i] );
		else if ( ( argv[i][0] != '/' ) && ( PrinterName[0] != '\0' ) && ( DriverName[0] == '\0' ) )
			wcscpy_s( DriverName, SHORT_BUF_LEN, argv[i] );

		else {
			_wcsupr_s( argv[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			fwprintf( stderr, L"Unknown option: \"%s\"\n", argv[i] );
			return 1;
		}

		i++;
	}


	if ( PrinterName[0] == '\0' ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"Missing printer name (/? for syntax help)\n" );
		return 1;
	}
	else if ( IsNetPrinterName( PrinterName ) ) {

		//  Connect to network printer:

		rv = 0;

		if ( ( DriverName[0] != '\0' ) || ( PortName[0] != '\0' ) || ( HostAddress[0] != '\0' ) || ( LPRQueueName[0] != '\0' ) || ( InfFilename[0] != '\0' ) || ( PtrComment[0] != '\0' ) || ( PtrLocation[0] != '\0' ) ) {
			DisplayErrorHeader();
			fwprintf( stderr, L"Don't combine network printer name with other options." );
			return 1;
		}

		if ( BeVerbose )
			wprintf( L"Network printer name: \"%s\" \n", PrinterName );

		if ( DoNothing )
			return 0;

		if ( CheckDisableRegistryKeys() )
			return 2;


		ConsoleShow( L"Connect to printer: " );

		if ( !AddPrinterConnection( PrinterName ) ) {

			rv = GetLastError();
			if ( !rv )
				rv = -1;

			ConsoleShow( NULL );
			DisplayErrorHeader();
			GetErrorText( rv, ErrorMessage );
			fwprintf( stderr, L"Failed to connect to printer \"%s\" (%s)\n", PrinterName, ErrorMessage );
			return 2;
		}

		if ( ChangePrefs )
			if ( SetPrinterPrefs( PrinterName, PrefColor, PrefDuplex, PrefOrient, PrefPaper, TRUE ) )
				rv = 2;
		
		if ( TestPage || MakeDefault ) {

			typedef int ( __stdcall *PRINTUIPROC )( HWND, HINSTANCE, LPTSTR, int );
			HMODULE hPrintUI = NULL;
			PRINTUIPROC PrintUIEntry = NULL;
			wchar_t commandbuf[MAX_ARG_LEN] = L"";


			hPrintUI = LoadLibrary( L"printui.dll" );

			if ( !hPrintUI ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Can't load PRINTUI.DLL\n" );
				return 2;
			}

			PrintUIEntry = (PRINTUIPROC) GetProcAddress( hPrintUI, "PrintUIEntryW" );

			if ( !PrintUIEntry ) {
				DisplayErrorHeader();
				fwprintf( stderr, L"Can't get address for PrintUIEntryW\n" );
				FreeLibrary( hPrintUI );
				return 2;
			}

			if ( TestPage ) {
				ConsoleShow( L"Send test page: " );

				wsprintf( commandbuf, L" /k /q /n \"%s\"", PrinterName );
				(PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );
			}

			if ( MakeDefault ) {
				ConsoleShow( L"Set default: " );
				SetDefaultPrinter( PrinterName );
			}
		}

		ConsoleShow( NULL );
		wprintf( L"Connected to printer \"%s\".\n", PrinterName );
		return rv;
	}


	if ( DriverName[0] == '\0' ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"Missing driver name\n" );
		return 1;
	}

	if ( ( PortName[0] == '\0' ) && ( HostAddress[0] == '\0' ) ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"Must specify either /A:address or /P:port\n" );
		return 1;
	}
	else if ( ( PortName[0] != '\0' ) && ( HostAddress[0] != '\0' ) ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"Specify only one of /A:address or /P:port\n" );
		return 1;
	}

	if ( ( HostAddress[0] == '\0' ) && ( LPRQueueName[0] != '\0' ) ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"Use /Q:queue in comjunction with /A:address\n" );
		return 1;
	}


	if ( HostAddress[0] != '\0' ) {
		//  Check for a TCP port number in HostAddress:

		int k = -1;
		for ( int j = 0; HostAddress[j] != '\0'; j++ )
			if ( HostAddress[j] == ':' )
				k = j;

		if ( k > 0 )
			if ( iswdigit( HostAddress[k + 1] ) ) {
				long int p = GetInt( HostAddress + k + 1 );

				if ( ( p >= 0 ) && ( p <= 0xffff ) ) {
					TCPPort = (WORD) p;
					TCPPortSpecified = TRUE;
					HostAddress[k] = '\0';
				}
			}

		HostNameIsDottedQuad = LooksLikeAnIP( HostAddress );
		LPRPrinterPort = ( ( LPRQueueName[0] != '\0' ) || ( TCPPortSpecified && ( TCPPort == 515 ) ) );

		if ( !TCPPortSpecified )
			TCPPort = LPRPrinterPort ? 515 : 9100;

		if ( LPRPrinterPort && ( LPRQueueName[0] == '\0' ) )
			wcscpy_s( LPRQueueName, SHORT_BUF_LEN, DefaultQueueName );

		//	GeneratePortName( HostAddress, PortName );
	}


	if ( wcslen( PrinterName ) > 220 ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"Printer name is %d characters max\n", 220 );
		return 1;
	}

	if ( wcslen( PortName ) >= MAX_PORTNAME_LEN ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"Port name is %d characters max\n", MAX_PORTNAME_LEN - 1 );
		return 1;
	}

	if ( wcslen( LPRQueueName ) >= MAX_QUEUENAME_LEN ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"LPR queue name is %d characters max\n", MAX_QUEUENAME_LEN - 1 );
		return 1;
	}

	if ( wcslen( PtrComment ) > 255 ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"Printer comment is %d characters max\n", 255 );
		return 1;
	}

	if ( wcslen( PtrLocation ) > 255 ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"Printer location is %d characters max\n", 255 );
		return 1;
	}


	if ( InfFilename[0] ) {
		ReplaceChar( InfFilename, '/', '\\' );
		InsertOSVersion( InfFilename );
	}

	TweakPortName( PortName );

	if ( DataFilename[0] ) {
		ReplaceChar( DataFilename, '/', '\\' );
		InsertOSVersion( DataFilename );
	}


	if ( BeVerbose ) {
		wprintf( L"Printer name:  \"%s\"\n", PrinterName );
		wprintf( L"Driver name:   \"%s\"\n", DriverName );
		wprintf( L"Port name:     \"%s\"\n", PortName );
		wprintf( L"Host address:  \"%s\"\n", HostAddress );
		wprintf( L"TCP port:      %d\n", TCPPort );
        wprintf( L"Protocol:      %s\n", LPRPrinterPort ? L"LPR" : L"Raw TCP" );
        wprintf( L"Queue name:    \"%s\"\n", LPRQueueName );
		wprintf( L"INF filename:  \"%s\"\n", InfFilename );
		wprintf( L"Comment:       \"%s\"\n", PtrComment );
		wprintf( L"Location:      \"%s\"\n", PtrLocation );
		wprintf( L"Data filename: \"%s\"\n\n", DataFilename );
	}

	if ( DoNothing )
		return 0;

	if ( DataFilename[0] ) {
		if ( !IsAFile( DataFilename ) ) {
			DisplayErrorHeader();
			fwprintf( stderr, L"File not found: \"%s\"\n", DataFilename );
			return 2;
		}
	}

	if ( CheckDisableRegistryKeys() )
		return 2;

	if ( NoOverwrite )
		if ( PrinterExists( PrinterName ) ) {
			wprintf( L"Printer \"%s\" already exists; exiting.\n", PrinterName );
			return 0;
		}

	if ( HostAddress[0] != '\0' ) {
		rv = CreateIPPort( PortName, HostNameIsDottedQuad ? NULL : HostAddress, HostNameIsDottedQuad ? HostAddress : NULL, LPRQueueName, LPRPrinterPort ? PROTOCOL_LPR_TYPE : PROTOCOL_RAWTCP_TYPE, TCPPort );

		if ( rv ) {
			DisplayErrorHeader();
			GetErrorText( rv, ErrorMessage );
			fwprintf( stderr, L"Error creating IP port \"%s\" (%s)\n", PortName, ErrorMessage );
			return 2;
		}
	}


	typedef int ( __stdcall *PRINTUIPROC )( HWND, HINSTANCE, LPTSTR, int );
	HMODULE hPrintUI = NULL;
	PRINTUIPROC PrintUIEntry = NULL;
	wchar_t commandbuf[MAX_ARG_LEN] = L"";



	hPrintUI = LoadLibrary( L"printui.dll" );

	if ( !hPrintUI ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"Can't load PRINTUI.DLL\n" );
		return 2;
	}

	PrintUIEntry = (PRINTUIPROC) GetProcAddress( hPrintUI, "PrintUIEntryW" );

	if ( !PrintUIEntry ) {
		DisplayErrorHeader();
		fwprintf( stderr, L"Can't get address for PrintUIEntryW\n" );
		FreeLibrary( hPrintUI );
		return 2;
	}

	if ( !NoOverwrite ) {
		ConsoleShow( L"Delete old printer: " );

		wsprintf( commandbuf, L" /dl /n \"%s\" /q", PrinterName );
		// wprintf( L"Command string: %s \n", commandbuf );
		(PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );
	}

	if ( InfFilename[0] != '\0' ) {

		ConsoleShow( L"Install driver: " );

		wsprintf( commandbuf, L" /ia /m \"%s\" /f \"%s\" /q", DriverName, InfFilename );
		// wprintf( L"Command string: %s \n", commandbuf );
		rv = (PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );

		if ( rv ) {
			ConsoleShow( NULL );
			DisplayErrorHeader();
			GetErrorText( rv, ErrorMessage );
			fwprintf( stderr, L"Error installing driver (%s)\n", ErrorMessage );
			FreeLibrary( hPrintUI );
			return 2;
		}
	}

	ConsoleShow( L"Create new printer: " );

	wsprintf( commandbuf, L" /if /b \"%s\" %s%s%s /q /r \"%s\" /m \"%s\" /z /u", PrinterName, ( ( InfFilename[0] == '\0' ) ? L"" : L"/f \"" ), ( ( InfFilename[0] == '\0' ) ? L"" : InfFilename ), ( ( InfFilename[0] == '\0' ) ? L"" : L"\" " ), PortName, DriverName );
	// wprintf( L"Command string: %s\n", commandbuf );

	rv = (PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );

	if ( !rv ) {

		//  Created the printer successfully!

		if ( DataFilename[0] ) {
			ConsoleShow( L"Restore data file: " );

			wsprintf( commandbuf, L" /Sr /q /n \"%s\" /a \"%s\" d g r", PrinterName, DataFilename );
			// wsprintf( L"Command string: %s\n", commandbuf );

			int rv2 = (PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );

			if ( rv2 ) {
				ConsoleShow( NULL );
				DisplayErrorHeader();
				GetErrorText( rv2, ErrorMessage );
				fwprintf( stderr, L"Error loading data file \"%s\":\n" );
				fwprintf( stderr, L"%s\n\n", ErrorMessage );
			}
		}

		if ( AdditionalSettings ) {
			ConsoleShow( L"Additional settings: " );

			wsprintf( commandbuf, L" /Xs /q /n \"%s\"", PrinterName );

			if ( PtrComment[0] != '\0' ) {
				wcscat_s( commandbuf, MAX_ARG_LEN, L" comment \"" );
				wcscat_s( commandbuf, MAX_ARG_LEN, PtrComment );
				wcscat_s( commandbuf, MAX_ARG_LEN, L"\"" );
			}

			if ( PtrLocation[0] != '\0' ) {
				wcscat_s( commandbuf, MAX_ARG_LEN, L" location \"" );
				wcscat_s( commandbuf, MAX_ARG_LEN, PtrLocation );
				wcscat_s( commandbuf, MAX_ARG_LEN, L"\"" );
			}
			
			// wprintf( L"Command string: %s\n", commandbuf );
			(PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );
		}

		if ( TestPage ) {
			ConsoleShow( L"Send test page: " );

			wsprintf( commandbuf, L" /k /q /n \"%s\"", PrinterName );
			(PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );
		}

		if ( MakeDefault ) {
			ConsoleShow( L"Set default: " );
			SetDefaultPrinter( PrinterName );
		}

	}

	FreeLibrary( hPrintUI );

	ConsoleShow( NULL );

	if ( rv ) {
		DisplayErrorHeader();
		GetErrorText( rv, ErrorMessage );
		fwprintf( stderr, L"Failed to create printer (%s)\n", ErrorMessage );
		return 2;
	}
	else {
		rv = 0;
		wprintf( L"Created printer \"%s\".\n", PrinterName );

		if ( ChangePrefs )
			if ( SetPrinterPrefs( PrinterName, PrefColor, PrefDuplex, PrefOrient, PrefPaper, TRUE ) )
				rv = 2;

		return rv;
	}
}
